# SPDX-FileCopyrightText: 2025 German Aerospace Center <amiris@dlr.de>
#
# SPDX-License-Identifier: Apache-2.0

"""Hosts the command to run AMIRIS."""

import os
from pathlib import Path

from amirispy.source.cli import RunOptions
from amirispy.source.exception import AMIRISError
from amirispy.source.fameio_calls import compile_output, call_amiris, compile_input, determine_all_paths
from amirispy.source.files import ensure_folder_accessible
from amirispy.source.java import check_java
from amirispy.source.logs import log_error

_ERR_NOT_A_FILE = "Specified path '{}' is no file."


def run_amiris(options: dict) -> None:
    """Compiles a scenario, then runs AMIRIS, and finally extracts results.

    Compiles scenario to protobuf using fameio.scripts.make_config,
    executes AMIRIS, and extracts results using fameio.scripts.convert_results

    Args:
        options: dictionary of command line instructions

    Raises:
        AMIRISError: if any error occurred during execution of AMIRIS; logged with level "ERROR"
    """
    check_java(skip=options[RunOptions.NO_CHECKS])
    path_to_scenario: Path = options[RunOptions.SCENARIO]
    if not path_to_scenario.is_file():
        raise log_error(AMIRISError(_ERR_NOT_A_FILE.format(path_to_scenario)))
    original_wd = Path.cwd()

    paths = determine_all_paths(path_to_scenario, original_wd, options)
    ensure_folder_accessible(paths["RESULT_FOLDER"])

    os.chdir(paths["SCENARIO_DIRECTORY"])
    compile_input(options, paths)
    os.chdir(paths["RESULT_FOLDER"])
    call_amiris(paths)
    compile_output(options, paths)
    os.remove(paths["INPUT_PB"])
    os.chdir(original_wd)



# SPDX-FileCopyrightText: 2025 German Aerospace Center <amiris@dlr.de>
#
# SPDX-License-Identifier: Apache-2.0

"""Hosts exceptions thrown by AMIRIS-Py."""


class AMIRISError(Exception):
    """An error occurred during the execution of AMIRIS-Py."""



# SPDX-FileCopyrightText: 2025 German Aerospace Center <amiris@dlr.de>
#
# SPDX-License-Identifier: Apache-2.0

"""Hosts command line parsing, its options and errors."""

import argparse
from enum import Enum, auto
from pathlib import Path
from typing import Any, Dict, Tuple, Optional, List

from amirispy.scripts.subcommands.download import DownloadMode
from amirispy.source.files import find_files
from amirispy.source.logs import LogLevels


class CommandLineError(Exception):
    """An error that occurred while parsing the command line options."""


AMIRIS_PARSER = "Command-line interface to the electricity market model AMIRIS"
AMIRIS_LOG_FILE_HELP = "Provide logging file (default: None)"
AMIRIS_LOG_LEVEL_HELP = f"Choose logging level (default: {LogLevels.ERROR.name})"
AMIRIS_COMMAND_HELP = "Choose one of the following commands:"

INSTALL_DEPRECATION = "'install' command is deprecated: use 'download' instead."
DOWNLOAD_HELP = "Downloads and extracts latest open access AMIRIS instance"
DOWNLOAD_URL_MODEL_HELP = "URL to download AMIRIS model from (default: latest AMIRIS artifact)"
DOWNLOAD_TARGET_HELP = "Folder to download 'amiris-core_<version>-jar-with-dependencies.jar' to (default: './')"
DOWNLOAD_FORCE_HELP = "Force download to overwrite existing AMIRIS and/or example files (default: False)"
DOWNLOAD_MODE_HELP = "Choose to download model and examples `all` (default), or `model` / `examples` only"

RUN_HELP = "Compile scenario, execute AMIRIS, and extract results"
RUN_JAR_HELP = "Path to 'amiris-core_<version>-jar-with-dependencies.jar (default: 'amiris*.jar' in current directory)"
RUN_JAR_DEFAULT = "amiris*.jar"
RUN_JAR_ERR_MISSING = "No jar-file found with pattern: '{}'"
RUN_JAR_ERR_TOO_MANY = "More than one jar-file found with pattern: '{}'"
RUN_SCENARIO_HELP = "Path to a scenario yaml-file"
RUN_OUTPUT_HELP = "Directory to write output to"
RUN_OUTPUT_OPTION_HELP = (
    "optional pass through of FAME-Io's output conversion options, see "
    "https://gitlab.com/fame-framework/fame-io/-/blob/main/README.md#read-fame-results"
)
RUN_NO_CHECK_HELP = "Skip checks for Java installation and correct version"

BATCH_HELP = "Batch mode to perform multiple runs each with scenario compilation, execution, and results extraction"
BATCH_SCENARIO_HELP = "Path to single or list of: scenario yaml-files or their enclosing directories"
BATCH_RECURSIVE_HELP = "Option to recursively search in provided Path for scenario (default: False)"
DEFAULT_PATTERN = "*.y*ml"
BATCH_PATTERN_HELP = f"Optional name pattern that scenario files searched for must match (default: '{DEFAULT_PATTERN}')"

COMPARE_HELP = "Compare if results of two AMIRIS runs and equivalent"
COMPARE_EXPECTED_HELP = "Path to folder with expected results"
COMPARE_TEST_HELP = "Path to folder with results to test"
COMPARE_IGNORE_HELP = "Optional list of file names to not be compared"
URL_LATEST_AMIRIS = "https://gitlab.com/dlr-ve/esy/amiris/amiris/-/jobs/artifacts/main/download?job=deploy:jdk11"


class GeneralOptions(Enum):
    """Specifies general options for workflow."""

    LOG = auto()
    LOGFILE = auto()


class Command(Enum):
    """Specifies command to execute."""

    RUN = auto()
    DOWNLOAD = auto()
    COMPARE = auto()
    BATCH = auto()


class CompareOptions(Enum):
    """Options for command `compare`."""

    EXPECTED = auto()
    TEST = auto()
    IGNORE = auto()


class DownloadOptions(Enum):
    """Options for command `download`."""

    URL = auto()
    TARGET = auto()
    FORCE = auto()
    MODE = auto()


class RunOptions(Enum):
    """Options for command `run`."""

    JAR = auto()
    SCENARIO = auto()
    OUTPUT = auto()
    OUTPUT_OPTIONS = auto()
    NO_CHECKS = auto()


class BatchOptions(Enum):
    """Options for command `batch`."""

    JAR = auto()
    SCENARIOS = auto()
    OUTPUT = auto()
    OUTPUT_OPTIONS = auto()
    NO_CHECKS = auto()
    RECURSIVE = auto()
    PATTERN = auto()


Options = {
    Command.COMPARE: CompareOptions,
    Command.RUN: RunOptions,
    Command.DOWNLOAD: DownloadOptions,
    Command.BATCH: BatchOptions,
}


def arg_handling_run(input_args: Optional[List[str]] = None) -> Tuple[Command, Dict[Enum, Any]]:
    """Handles command line arguments for `amiris` and returns `command` and its options `args`.

    Allows to set args from a list of input_args.

    Raises:
        CommandLineError: if a custom error occurred during command-line handling
    """

    parent_parser = argparse.ArgumentParser(prog="amiris", description=AMIRIS_PARSER)
    parent_parser.add_argument("-lf", "--logfile", type=Path, required=False, help=AMIRIS_LOG_FILE_HELP)
    parent_parser.add_argument(
        "-l",
        "--log",
        default=LogLevels.WARN.name,
        choices=[level.name.lower() for level in LogLevels],
        help=AMIRIS_LOG_LEVEL_HELP,
    )
    subparsers = parent_parser.add_subparsers(dest="command", required=True, help=AMIRIS_COMMAND_HELP)

    download_parser = subparsers.add_parser("download", help=DOWNLOAD_HELP)
    add_download_arguments(download_parser)

    install_parser = subparsers.add_parser("install", help=INSTALL_DEPRECATION)
    add_download_arguments(install_parser)

    run_parser = subparsers.add_parser("run", help=RUN_HELP)
    run_parser.add_argument("--scenario", "-s", type=Path, required=True, help=RUN_SCENARIO_HELP)
    run_parser.add_argument("--jar", "-j", type=str, default=RUN_JAR_DEFAULT, help=RUN_JAR_HELP)
    run_parser.add_argument("--output", "-o", type=Path, default=Path("./result"), help=RUN_OUTPUT_HELP)
    run_parser.add_argument("--output-options", "-oo", type=str, default="", help=RUN_OUTPUT_OPTION_HELP)
    run_parser.add_argument("--no-checks", "-nc", action="store_true", default=False, help=RUN_NO_CHECK_HELP)

    batch_parser = subparsers.add_parser("batch", help=BATCH_HELP)
    batch_parser.add_argument("--scenarios", "-s", nargs="+", type=Path, required=True, help=BATCH_SCENARIO_HELP)
    batch_parser.add_argument("--jar", "-j", type=str, default=RUN_JAR_DEFAULT, help=RUN_JAR_HELP)
    batch_parser.add_argument("--output", "-o", type=Path, default=Path("./result"), help=RUN_OUTPUT_HELP)
    batch_parser.add_argument("--recursive", "-r", default=False, action="store_true", help=BATCH_RECURSIVE_HELP)
    batch_parser.add_argument("--pattern", "-p", type=str, default=DEFAULT_PATTERN, help=BATCH_PATTERN_HELP)
    batch_parser.add_argument("--output-options", "-oo", type=str, default="", help=RUN_OUTPUT_OPTION_HELP)
    batch_parser.add_argument("--no-checks", "-nc", action="store_true", default=False, help=RUN_NO_CHECK_HELP)

    compare_parser = subparsers.add_parser("compare", help=COMPARE_HELP)
    compare_parser.add_argument("--expected", "-e", type=Path, required=True, help=COMPARE_EXPECTED_HELP)
    compare_parser.add_argument("--test", "-t", type=Path, required=True, help=COMPARE_TEST_HELP)
    compare_parser.add_argument("--ignore", "-i", required=False, help=COMPARE_IGNORE_HELP)

    args = vars(parent_parser.parse_args(input_args))
    command = get_command_from_string(args.pop("command"))
    if command in (Command.RUN, Command.BATCH):
        args["jar"] = find_single_executable_path(args["jar"])
    return command, enumify(command, args)


def add_download_arguments(parser: argparse.ArgumentParser) -> None:
    """Adds download arguments to given parser."""
    parser.add_argument("--url", "-u", default=URL_LATEST_AMIRIS, help=DOWNLOAD_URL_MODEL_HELP)
    parser.add_argument("--target", "-t", type=Path, default=Path("./"), help=DOWNLOAD_TARGET_HELP)
    parser.add_argument("--force", "-f", default=False, action="store_true", help=DOWNLOAD_FORCE_HELP)
    parser.add_argument(
        "--mode",
        "-m",
        type=str.lower,  # noqa
        choices=[mode.name.lower() for mode in DownloadMode],
        default=DownloadMode.ALL.name,
        help=DOWNLOAD_MODE_HELP,
    )


def get_command_from_string(command_str: str) -> Command:
    """Returns Command extracted from given string."""
    if command_str == "install":
        print(INSTALL_DEPRECATION)
        command_str = "download"
    return Command[command_str.upper()]


def find_single_executable_path(pattern: str) -> Path:
    """Searches for single file with given `pattern`; Raises an error if no file or more than one file match."""
    matching_files = find_files(pattern)
    if len(matching_files) == 0:
        raise CommandLineError(RUN_JAR_ERR_MISSING.format(pattern))
    if len(matching_files) > 1:
        raise CommandLineError(RUN_JAR_ERR_TOO_MANY.format(pattern))
    return Path(matching_files[0])


def enumify(command: Command, args: dict) -> Dict[Enum, Any]:
    """Matches `args` for given `command` to their respective Enum."""

    result = {}
    for option in GeneralOptions:
        result[option] = args.pop(option.name.lower())

    for option in Options[command]:
        result[option] = args.pop(option.name.lower())
    return result



# SPDX-FileCopyrightText: 2025 German Aerospace Center <amiris@dlr.de>
#
# SPDX-License-Identifier: Apache-2.0

"""Holds methods to check installed version of Java."""

import logging as log
import re
import shutil
import subprocess

from amirispy.source.exception import AMIRISError
from amirispy.source.logs import log_critical

_ERR_NO_JAVA = "No Java installation found. See {} for further instructions."
_ERR_JAVA_VERSION = "Local Java version '{}' does not match requirements '>{}'."
_URL_INSTALLATION_INSTRUCTIONS = "https://gitlab.com/dlr-ve/esy/amiris/amiris-py#further-requirements"

JAVA_VERSION_PATTERN = '"(\d+\.\d+).*"'  # noqa
JAVA_VERSION_MINIMUM = 11


def check_java_installation(raise_exception: bool = False) -> None:
    """Checks if the java command is available.

    Args:
        raise_exception: if True, an Exception is raised, else a warning

    Raises:
        AMIRISError: if Java installation is not found; logged with level "WARNING" (default) or "CRITICAL"
    """
    if not shutil.which("java"):
        if raise_exception:
            raise log_critical(AMIRISError(_ERR_NO_JAVA.format(_URL_INSTALLATION_INSTRUCTIONS)))
        log.warning(_ERR_NO_JAVA.format(_URL_INSTALLATION_INSTRUCTIONS))


def check_java_version(raise_exception: bool = False) -> None:
    """Checks if Java version is compatible with requirements of FAME.

    Args:
        raise_exception: if True, an Exception is raised, else a warning

     Raises:
        AMIRISError: if Java version is not compatible; logged with level "WARNING" (default) or "CRITICAL"
    """
    version_raw = subprocess.check_output(["java", "-version"], stderr=subprocess.STDOUT)
    version_number = re.search(JAVA_VERSION_PATTERN, str(version_raw)).groups()[0]

    if float(version_number) < JAVA_VERSION_MINIMUM:
        if raise_exception:
            raise log_critical(AMIRISError(_ERR_JAVA_VERSION.format(version_number, JAVA_VERSION_MINIMUM)))
        log.warning(_ERR_JAVA_VERSION.format(version_number, JAVA_VERSION_MINIMUM))


def check_java(skip: bool) -> None:
    """Checks both Java installation and version if not `skip`.

    Args:
        skip: if enabled, checks are skipped

    Raises:
          AMIRISError: if any check fails; logged with level "CRITICAL"
    """
    if not skip:
        check_java_installation(raise_exception=True)
        check_java_version(raise_exception=True)



# SPDX-FileCopyrightText: 2025 German Aerospace Center <amiris@dlr.de>
#
# SPDX-License-Identifier: CC0-1.0

"""Contains the main script and its subcommands called by CLI."""

from .amiris import amiris_cli
from ..source.exception import AMIRISError


def amiris():
    try:
        amiris_cli()
    except AMIRISError as e:
        raise SystemExit(1) from e



# SPDX-FileCopyrightText: 2025 German Aerospace Center <amiris@dlr.de>
#
# SPDX-License-Identifier: Apache-2.0

"""Hosts methods to search and check files or folders."""

import logging
import os
from pathlib import Path
from typing import List

from amirispy.source.exception import AMIRISError
from amirispy.source.logs import log_error

_CSV_FILE_ENDING = ".csv"

_ERR_NOT_A_FOLDER = "Given Path '{}' is not a directory."
_ERR_MISSING_FOLDER = "Specified directory '{}' is missing."
_ERR_NO_ACCESS = "No writing permission to directory '{}'."

_WARN_NOT_EMPTY = "Folder '{}' is not empty - overriding files."


def get_all_csv_files_in_folder_except(folder: Path, exceptions: List[str] = None) -> List[Path]:
    """Find all csv files in a folder that can optionally ignore a files with a given file name.

    Args:
        folder: to search for csv files - file ending is **not** case-sensitive
        exceptions: optional, files names (without file ending) listed here will be ignored - **not** case-sensitive

    Returns:
        Full file Paths for files ending with ".csv" not listed in exceptions

    Raises:
        AMIRISError: if folder is missing; logged with level "ERROR"
    """
    if not folder.is_dir():
        raise log_error(AMIRISError(_ERR_MISSING_FOLDER.format(folder)))

    if exceptions is None:
        exceptions = list()
    exceptions = [item.upper() for item in exceptions]
    all_csvs = [file for file in folder.glob(f"*{_CSV_FILE_ENDING}")]
    return [file for file in all_csvs if file.stem not in exceptions]


def ensure_folder_accessible(result_folder) -> None:
    """Ensure that given folder exists and can be written to.

    Raises:
        AMIRISError: if any error occurred during execution of AMIRIS; logged with level "ERROR"
    """
    ensure_folder_exists(result_folder)
    check_if_write_access(result_folder)
    warn_if_not_empty(result_folder)


def ensure_folder_exists(path: Path) -> None:
    """Returns Path to a directory and creates the folder if required.

    If given Path is an existing folder: does nothing, else creates new folder (including parent folders)

    Args:
        path: to check and create if not existing

    Raises:
        AMIRISError: if path not found or an existing file; logged with level "ERROR"
    """
    try:
        if path.is_file():
            raise log_error(AMIRISError(_ERR_NOT_A_FOLDER.format(path)))
        if not path.is_dir():
            path.mkdir(parents=True)
    except FileNotFoundError:
        raise log_error(AMIRISError(_ERR_NOT_A_FOLDER.format(path)))


def check_if_write_access(path: Path) -> None:
    """Checks writing access in given `path` by writing and deleting a temporary file.

    Args:
        path: to check if writing access

    Raises:
        AMIRISError: if no writing access in given path; logged with level "ERROR"
    """
    try:
        tmp_file_name = Path(path, "amirispy_write_access_test_file")
        file = open(tmp_file_name, "w")
        file.close()
        os.remove(tmp_file_name)
    except (OSError, IOError):
        raise log_error(AMIRISError(_ERR_NO_ACCESS.format(path)))


def warn_if_not_empty(folder: Path) -> None:
    """Logs a warning if given folder is not empty.

    Args:
        folder: to check for files
    """
    if list(folder.glob("*")):
        logging.warning(_WARN_NOT_EMPTY.format(folder))


def ensure_absolute(target: Path, base: Path) -> Path:
    """Returns absolute path for given target - relative to base if target is relative.

    Returns given `target` path if it is absolute.
    If `target` is relative, return an absolute path, interpreting `target` relative to the provided `base` path.

    Args:
        target: absolute path, or interpreted as relative to base path
        base: path target is interpreted to if target is not absolute itself
    """
    return target if target.is_absolute() else base.joinpath(target)


def find_files(pattern: str) -> list[Path]:
    """Searches and returns files that match given search pattern.

    Args:
        pattern: a relative or absolute file path or search pattern

    Returns:
        list of file paths that match the provided pattern
    """
    pattern_path = Path(pattern)
    if pattern_path.is_absolute():
        search_dir, glob_pattern = pattern_path.parent, pattern_path.name
    else:
        search_dir, glob_pattern = Path("."), pattern
    return sorted(search_dir.glob(glob_pattern))



# SPDX-FileCopyrightText: 2025 German Aerospace Center <amiris@dlr.de>
#
# SPDX-License-Identifier: Apache-2.0

"""Hosts the command to batch-run AMIRIS."""

import logging as log
import os
from pathlib import Path
from typing import List

from fameio.input import InputError
from fameio.input.loader import load_yaml
from fameio.input.scenario import Scenario
from fameio.input.validator import SchemaValidator
from fameio.logs import fameio_logger, LogLevel

from amirispy.scripts.subcommands.run import run_amiris
from amirispy.source.cli import BatchOptions, RunOptions
from amirispy.source.exception import AMIRISError
from amirispy.source.files import ensure_absolute
from amirispy.source.java import check_java
from amirispy.source.logs import log_and_print, log_error

_ERR_ALL_PATHS_INVALID = "Provided scenario path(s) '{}' contain no valid scenario YAML files."
_ERR_ABS_PATTERN = "Provided pattern '{}' must not start with a '/'."
_WARN_PATH_NOT_EXISTING = "Provided path '{}' is ignored as it is neither a file nor a directory."
_WARN_NO_VALID_FAME_SCENARIO = "'{}' is not a valid scenario file. You may improve the file name pattern using `-p`"
_PRINT_RUN = "AMIRIS run {} of {}: Simulate file '{}', find results at '{}'"


def batch_run_amiris(options: dict) -> None:
    """Compile, run, and extract multiple scenarios.

    Compile multiple scenarios to protobuf using fameio.scripts.make_config, execute AMIRIS,
    and extract results using fameio.scripts.convert_results.

    Args:
        options: dictionary of command line instructions

    Raises:
        AMIRISError: if any error occurs; logged with level "ERROR"
    """
    check_java(skip=options[BatchOptions.NO_CHECKS])
    input_yaml_files = find_valid_scenarios(
        options[BatchOptions.SCENARIOS], options[BatchOptions.RECURSIVE], options[BatchOptions.PATTERN]
    )
    run_options = options.copy()
    run_options.update(
        {
            RunOptions.JAR: options[BatchOptions.JAR],
            RunOptions.NO_CHECKS: True,
            RunOptions.OUTPUT_OPTIONS: options[BatchOptions.OUTPUT_OPTIONS],
        }
    )

    for i, input_yaml_file in enumerate(input_yaml_files):
        output_folder = ensure_absolute(options[BatchOptions.OUTPUT].joinpath(str(i)), Path.cwd())
        log_and_print(_PRINT_RUN.format(i + 1, len(input_yaml_files), input_yaml_file, output_folder))
        run_options.update({RunOptions.SCENARIO: input_yaml_file, RunOptions.OUTPUT: output_folder})
        run_amiris(run_options)


def find_valid_scenarios(search_paths: List[Path], recursive: bool, pattern: str) -> List[Path]:
    """Searches for valid scenario YAML files in given `input_yaml_paths`.

    Args:
        search_paths: path(s) which are to be searched for valid scenario files
        recursive: if true, subdirectories of each search path are searched as well
        pattern: that file names must match to be returned

    Returns:
        List of Paths to valid scenario files

    Raises:
        AMIRISError: if none of the files in path are a valid scenario or pattern is invalid; logged with level "ERROR"
    """
    files_to_test = get_inner_yaml_files(search_paths, recursive, pattern)
    fameio_logger(LogLevel.PRINT.name)
    scenario_files = [file for file in files_to_test if is_valid_fame_input_yaml(file)]
    if not scenario_files:
        raise log_error(AMIRISError(_ERR_ALL_PATHS_INVALID.format(search_paths)))
    log.info(f"Found these scenario file(s) '{scenario_files}'.")
    return scenario_files


def get_inner_yaml_files(paths: List[Path], recursive: bool, pattern: str) -> List[Path]:
    """Returns a list of all YAML files in `paths` (and all subdirectories when `recursive` is set to True).

    Args:
        paths: to search for YAML files
        recursive: if True, subdirectories are also searched
        pattern: that file names must match to be returned

    Returns:
        List of Paths to YAML files contained in `paths` (and its subdirectories)

    Raises:
        AMIRISError: if pattern is invalid; logged with level "ERROR"
    """
    yaml_files = []
    for path in paths:
        if path.is_file():
            yaml_files.append(path)
        elif path.is_dir():
            try:
                yaml_files.extend([f for f in path.glob(pattern) if f.is_file()])
            except NotImplementedError as e:
                raise log_error(AMIRISError(_ERR_ABS_PATTERN.format(pattern))) from e
            if recursive:
                yaml_files.extend(get_inner_yaml_files([f for f in path.glob("*") if f.is_dir()], recursive, pattern))
        else:
            log.warning(_WARN_PATH_NOT_EXISTING.format(path))
    return yaml_files


def is_valid_fame_input_yaml(file_to_test: Path) -> bool:
    """Checks if the given path points to a valid FAME input YAML file. If not, an error is logged.

    Args:
        file_to_test: Path to a file

    Returns:
        True if the given path points to valid FAME input YAML file, otherwise False
    """
    file_path = ensure_absolute(file_to_test, Path.cwd())
    original_cwd = Path.cwd()
    try:
        os.chdir(file_path.parents[0])
        scenario = Scenario.from_dict(load_yaml(file_path))
        SchemaValidator.validate_scenario_and_timeseries(scenario)
        return True
    except InputError as e:
        log.warning(_WARN_NO_VALID_FAME_SCENARIO.format(file_to_test))
        log.info(f"Error: {e}")
    finally:
        os.chdir(original_cwd)
    return False



# SPDX-FileCopyrightText: 2025 German Aerospace Center <amiris@dlr.de>
#
# SPDX-License-Identifier: Apache-2.0

"""Hosts the command to download the AMIRIS executable and example data."""

import logging as log
import os
import shutil
import zipfile
from enum import Enum, auto
from pathlib import Path
from urllib.error import HTTPError

import wget

from amirispy.source.exception import AMIRISError
from amirispy.source.files import ensure_folder_exists, check_if_write_access
from amirispy.source.logs import log_and_print, log_critical, log_error
from amirispy.source.java import check_java_installation, check_java_version

MSG_SKIPPED_DOWNLOAD = "Skipped download of {} due to option '-m/--mode {}'"
URL_EXAMPLES = "https://gitlab.com/dlr-ve/esy/amiris/examples/-/archive/main/examples-main.zip"

DEPLOY_URL = "https://gitlab.com/dlr-ve/esy/amiris/amiris/-/artifacts"
ERR_URL_INVALID = "Download failed with error: '{}'. Please find latest deploy artifacts here: {}."
ERR_FILE_EXISTS = "'{}' already exists in '{}'. Use `-f/--force` to override anyway."
ERR_NOT_A_ZIP_FILE = "Downloaded file is not a zip file: Could not unzip."


class DownloadMode(Enum):
    """Download `mode` options."""

    ALL = auto()
    MODEL = auto()
    EXAMPLES = auto()


def download_amiris(url: str, target_folder: Path, force: bool, mode: str) -> None:
    """Download and unzip AMIRIS from given url, overwriting existing AMIRIS file if `force` is enabled.

    Args:
        url: where to download packaged AMIRIS file from
        target_folder: folder where to save the AMIRIS to
        force: flag to overwrite existing AMIRIS downloads of same version and existing examples
        mode: mode of download (all, model only, examples only)

    Raises:
        AMIRISError: if any check fails, logged with level "WARNING"
    """

    ensure_folder_exists(target_folder)
    check_if_write_access(target_folder)

    mode = DownloadMode[mode.upper()]
    if mode == DownloadMode.ALL or mode == DownloadMode.MODEL:
        _conduct_model_download(url, target_folder, force)
    else:
        log.info(MSG_SKIPPED_DOWNLOAD.format("AMIRIS model", mode.name.lower()))

    if mode == DownloadMode.ALL or mode == DownloadMode.EXAMPLES:
        _conduct_example_download(target_folder, force)
    else:
        log.info(MSG_SKIPPED_DOWNLOAD.format("examples", mode.name.lower()))

    check_java_installation()
    check_java_version()


def _conduct_model_download(url: str, target_folder: Path, force: bool) -> None:
    """Downloads and extracts AMIRIS from `url` to `target_folder`. Overwrites existing files if `force` is True.

    Args:
        url: where to download packaged AMIRIS file from
        target_folder: folder where to save the AMIRIS to
        force: flag to overwrite existing AMIRIS files

    Raises:
        AMIRISError: if download is not successful; logged with level "CRITICAL"
    """
    download_file_path = Path(target_folder, "amiris.zip")
    log.info("Starting download of AMIRIS")
    try:
        wget.download(url=url, out=str(download_file_path), bar=progress_bar)
    except HTTPError as e:
        raise log_critical(AMIRISError(ERR_URL_INVALID.format(e, DEPLOY_URL)))
    log.info(f"Downloaded file to '{download_file_path}'")
    _unzip(download_file_path, force, target_folder)
    log_and_print(f"AMIRIS download to '{target_folder}' completed.")


def _unzip(download_file_path, force, target_folder) -> None:
    """Unzips downloaded file.

    Args:
        download_file_path: path to zip file
        target_folder: folder where to save the AMIRIS to
        force: flag to overwrite existing AMIRIS files

    Raises:
        AMIRISError: if unzipping is not successful; logged with level "ERROR"
    """
    if zipfile.is_zipfile(download_file_path):
        with zipfile.ZipFile(download_file_path, "r") as zip_ref:
            zip_ref.extractall(target_folder)
        os.remove(download_file_path)

        if force:
            for file in target_folder.glob("target/*with-dependencies.jar"):
                shutil.copy(src=str(file), dst=target_folder)
            log.info(f"Unzipped file content to '{target_folder}'")
        else:
            for file in target_folder.glob("target/*with-dependencies.jar"):
                try:
                    shutil.move(src=str(file), dst=target_folder)
                except shutil.Error:
                    log.error(ERR_FILE_EXISTS.format(file.name, target_folder))
        shutil.rmtree(Path(target_folder, "target"))
    else:
        raise log_error(AMIRISError(ERR_NOT_A_ZIP_FILE))


def progress_bar(current, total, _) -> str:
    """Progress bar that adds a newline in the end."""
    progress_message = "Download status: "
    if total < 0:
        progress_message += "unknown"
    elif current < total:
        progress_message += "%d%% [%d / %d] bytes" % (current / total * 100, current, total)
    else:
        progress_message += "done\n"
    return progress_message


def _conduct_example_download(target_folder: Path, force: bool) -> None:
    """Downloads and extracts examples for AMIRIS to `target_folder`. Overwrites existing examples if `force` is True.

    Args:
        target_folder: folder where to save the examples to
        force: if True, existing examples are overwritten
    """
    download_file_path = Path(target_folder, "examples.zip")
    log.info("Starting download of examples")
    wget.download(url=URL_EXAMPLES, out=str(download_file_path), bar=progress_bar)
    print("")  # fix broken progress bar that misses the newline
    log.info(f"Downloaded examples to '{download_file_path}'")

    if Path(target_folder, "examples").exists() and not force:
        log.error(f"'examples' already exists in '{target_folder}'. Use `-f/--force` to override anyway.")
    else:
        if zipfile.is_zipfile(download_file_path):
            with zipfile.ZipFile(download_file_path, "r") as zip_ref:
                zip_ref.extractall(target_folder)
        else:
            log.info("Downloaded file is not a zip file: Could not unzip")
        shutil.move(src=f"{target_folder}/examples-main", dst=f"{target_folder}/examples")

        log_and_print(f"Examples download to '{target_folder}/examples' completed.")
    os.remove(download_file_path)



# SPDX-FileCopyrightText: 2025 German Aerospace Center <amiris@dlr.de>
#
# SPDX-License-Identifier: Apache-2.0

"""Hosts methods to call FAME-Io."""

import logging as log
import subprocess
from pathlib import Path
from typing import Dict

from fameio.cli.convert_results import handle_args as handle_output_args
from fameio.cli.options import Options, ResolveOptions, TimeOptions
from fameio.scripts import ScriptError
from fameio.scripts.convert_results import run as convert_results
from fameio.scripts.make_config import run as make_config

from amirispy.source.cli import GeneralOptions, RunOptions
from amirispy.source.exception import AMIRISError
from amirispy.source.files import ensure_absolute
from amirispy.source.logs import log_error

ERR_FAMEIO_OUTPUT_ARGS = "Invalid fameio output conversion options: '{}'"
ERR_COMPILE = "Error during input file compilation in FAME-Io from scenario `{}`"
ERR_RUN = "Error during AMIRIS model run. Call was '{}'"
ERR_CONVERT = "Error during converting output file in FAME-Io"
INFO_START_COMPILE = "Creating binary protobuf input file: `{}` from scenario `{}`"


def determine_all_paths(scenario_yaml: Path, working_directory: Path, options: Dict) -> Dict[str, Path]:
    """Determines all Paths for the given scenario file and working directory.

    Args:
        scenario_yaml: path to the scenario file
        working_directory: the base working directory
        options: amiris-py command line options for the `run` command

    Returns:
        Dictionary containing all Paths to final results and temporary files
    """
    result_folder = ensure_absolute(options[RunOptions.OUTPUT], working_directory)
    jar_file = ensure_absolute(Path(options[RunOptions.JAR]), working_directory)
    return {
        "SCENARIO_FILE": ensure_absolute(scenario_yaml, working_directory),
        "SCENARIO_DIRECTORY": scenario_yaml.parent,
        "BASE_NAME": scenario_yaml.stem,
        "JAR_FILE": jar_file,
        "SETUP_FILE": jar_file.parents[0].joinpath("fameSetup.yaml"),
        "RESULT_FOLDER": result_folder,
        "INPUT_PB": result_folder.joinpath("input.pb"),
        "OUTPUT_PB": result_folder.joinpath("output.pb"),
    }


def compile_input(options: dict, paths: Dict[str, Path]) -> None:
    """Creates protobuf file using given `options` and `paths`.

    Args:
        options: for logging
        paths: to given input and output files

    Raises:
        AMIRISError: if input protobuf fails to compile; logged with level "ERROR"
    """
    fameio_input_config = {
        Options.FILE: paths["SCENARIO_FILE"].resolve(),
        Options.LOG_LEVEL: options[GeneralOptions.LOG],
        Options.LOG_FILE: options[GeneralOptions.LOGFILE],
        Options.OUTPUT: paths["INPUT_PB"],
    }
    log.info(INFO_START_COMPILE.format(fameio_input_config[Options.OUTPUT], fameio_input_config[Options.FILE]))
    try:
        make_config(fameio_input_config)
    except ScriptError:
        raise log_error(AMIRISError(ERR_COMPILE.format(fameio_input_config[Options.FILE])))


def call_amiris(paths: Dict[str, Path]) -> None:
    """Run AMIRIS using given paths.

    Args:
        paths: to all required files

    Raises:
        AMIRISError: if any error occurred during AMIRIS model run; logged with level "ERROR"
    """
    call = 'java -jar "{}" -f "{}" -o "output.pb"'.format(paths["JAR_FILE"], paths["INPUT_PB"])
    log.info(f"Calling AMIRIS with {paths['SCENARIO_FILE']}")
    try:
        subprocess.run(call, shell=True, check=True)
    except subprocess.SubprocessError:
        raise log_error(AMIRISError(ERR_RUN.format(call)))


def compile_output(options: Dict, paths: Dict[str, Path]) -> None:
    """Conducts to conversion of AMIRIS output to CSV files based on given options and paths.

    Args:
        options: for logging
        paths: to given input and output files

    Raises:
        AMIRISError: if an error occurred during output file conversion; logged with level "ERROR"
    """
    default_fameio_output_config = {
        Options.FILE: paths["OUTPUT_PB"].resolve(),
        Options.LOG_LEVEL: options[GeneralOptions.LOG],
        Options.LOG_FILE: options[GeneralOptions.LOGFILE],
        Options.AGENT_LIST: None,
        Options.OUTPUT: ".",
        Options.SINGLE_AGENT_EXPORT: False,
        Options.MEMORY_SAVING: False,
        Options.RESOLVE_COMPLEX_FIELD: ResolveOptions.SPLIT.name,
        Options.TIME: TimeOptions.UTC.name,
        Options.TIME_MERGING: [0, 1800, 1799],
    }
    log.info("Converting protobuf to csv files")
    fameio_output_config = override_eligible_defaults(default_fameio_output_config, options[RunOptions.OUTPUT_OPTIONS])
    try:
        convert_results(fameio_output_config)
    except ScriptError:
        raise log_error(AMIRISError(ERR_CONVERT))


def override_eligible_defaults(defaults: Dict, overrides: str) -> Dict:
    """Replaces default fameio arguments with given overrides; ensures that the output folder and file are not modified.

    Args:
        defaults: standard fameio output options
        overrides: (possibly empty) list of fameio output arguments to be used instead

    Returns:
        dict of (modified) fameio arguments

    Raises:
        AMIRISError: if invalid fameio output conversion options are received; logged with level "ERROR"
    """
    fixed_output_folder = defaults[Options.OUTPUT]
    fixed_file_to_convert = defaults[Options.FILE]
    try:
        argument_list = overrides.split()
        fameio_output_config = handle_output_args(argument_list, defaults)
    except SystemExit:
        raise log_error(AMIRISError(ERR_FAMEIO_OUTPUT_ARGS.format(overrides)))
    fameio_output_config[Options.OUTPUT] = fixed_output_folder
    fameio_output_config[Options.FILE] = fixed_file_to_convert
    return fameio_output_config
